package datasink;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.net.*;

import datarecord.*;

/**
 * DataSink: abstract base class for components that send
 * data somewhere through write() calls by their client.
 *
 * The client's thread is used to put the offered data into
 * a buffer (when it invokes the write() method).
 * A DataSink contains a thread of its own that sits in a loop
 * reading from the buffer and writing to the DataSink's destination.
 * 
 * The write() interface operates on byte[]'s -- that's what the
 * client provides each write() call.
 * The DataSink defines a maximum number of bytes its willing to take
 * in one write() invocation.  (If the client wants to send more than
 * that, it has to make repeated write() invocations.)
 * That maximum size reflects the size of an individual DataRecord
 * object the DataSink uses to buffer the client's data.
 *
 * write() is non-blocking when there is room in the DataSink's buffer
 * to add a new DataRecord.  Otherwise, the client thread is blocked.
 * 
 * @version $Id: DataSink.java,v 1.10 2010/01/07 21:18:42 zahorjan Exp $
 */

abstract public class DataSink {

  /**
   * The data buffer between the local, writer thread and the client
   * thread.  (The buffer is accessed by client via the write() method.)
   */
  private ArrayBlockingQueue<DataRecord> dataBuffer;

  /**
   * An object that knows how to create DataRecord's.
   */
  private DataRecordManager recordManager;

  /**
   * Size of dataBuffer;
   */
  private static final int BUFFER_SIZE = 20;

  /**
   * Maximum number of bytes we're willing to take in one write()
   * invocation.  (Also the capacity of each DataRecord in the buffer.)
   *
   */
  private static final int MAX_RECORD_SIZE = 1500;

  /**
   * The sender thread (takes packets from the buffer and sends them off somewhere).
   */
  private Thread sender;

  /**
   * We need to keep track of what state the input source is in.
   * <ul>
   * <li>STARTING: object created but not yet connected to its own input
   * <li>CONNECTED: object created and connected to its input
   * <li>CLOSE_WAIT: object's source is out of data, but there may still data in the data
   *     waiting to be sent
   * <li>CLOSED: out of data.  Further read() invocations are errors.
   */
  public final static int STARTING = 0;
  public final static int CONNECTED = 1;
  public final static int CLOSE_WAIT = 2;
  public final static int CLOSED = 3;

  protected int state = STARTING;

  /**
   * The SendHelper implemented here depends on this abstract method
   * being implemented by the instantiated subclass.  That method should
   * take the provided bytes and send them on their way.
   */
  protected abstract void sendToDest(byte[] buf, int len) throws IOException;

  /**
   * close() has to block the caller until the sending thread has finished
   * all its writing and committed to terminating.  We use this semaphore
   * to effect the needed synchronization.
   */
  Semaphore closeSync;

  /**
   * All derived class objects must be capable of reinitializing themselves.
   */
  public abstract void reInit() throws FileNotFoundException, IOException;

  /** 
   * Helper class to implement sendding thread.  Loops trying to
   * emit previously buffered data.
   */
  private class SendHelper extends Thread {

    public void run() {

      try {
        while ( true ) {
          // Fetch next record from buffer.
          DataRecord record = dataBuffer.take();
          int recLength = record.length;
          sendToDest( record.data, recLength );
          recordManager.freeRecord( record );

          // all done?
          if ( recLength < 0 ) break;
        }
      }
      catch ( Exception e ) {
        // any of a large variety of IO related exceptions
        // might have occurred.
        // Just terminate this thread
        System.err.println("DataSink SendHelper exception: " + e.getMessage() );
        e.printStackTrace();
      }
      state = CLOSE_WAIT;
      closeSync.release();

    }
  }

  /** 
   * Creates a new DataSink. Protected because only subclasses should be instantiated.
   */
  protected DataSink() {
    // create buffer to hold outgoing packets
    dataBuffer = new ArrayBlockingQueue<DataRecord>(BUFFER_SIZE);
    // create the DataRecord allocator/cache
    recordManager = new DataRecordManager();
    // initialize the close synchronization semaphore
    closeSync = new Semaphore(0);
  }

  /**
   * Protected function to initialize (called by subclass constructor
   * when it has finished initializing).
   */
  protected void readyToWrite() {
    state = CONNECTED;
    dataBuffer.clear();
    // Start the sending thread.
    sender = this.new SendHelper();
    sender.start(); // new thread executes the run() method above; this thread simply returns
  }

  /**
   * Well... write.  Eventually.
   */

  public void write(byte[] buf, int length) throws IOException, InterruptedException {
    if ( state==CLOSED)
      throw new IOException("Attempt to write when state is " + (state==CLOSED?"CLOSED":"CLOSE_WAIT"));

    if ( length > MAX_RECORD_SIZE) 
      throw new IOException("DataSink.write(): length " + length + " exceeds maximum " + MAX_RECORD_SIZE);

    if ( length < 0 )
      throw new IOException("DataSink.write(): negative length: " + length );

    // why allocate the max size?  It's a more accurate
    // reflection of what is typically done when worried
    // about speed (i.e., using an array of DataRecords
    // for the buffer, allocated once, rather than allocating
    // new records all the time).
    DataRecord record = recordManager.getNewRecord(MAX_RECORD_SIZE);

    System.arraycopy(buf, 0, record.data, 0, length);  // Why copy the data?  The user might change buf[] 
                                                       // before we've managed to send it!
    record.length = length;
    dataBuffer.put(record);
  }

  /**
   * The user is promising never to send data again.  (We need this notification
   * to know when to terminate the sender thread.)
   */

  public void close() {
    state = CLOSE_WAIT;
    // put a length -1 data record in the buffer to indicate end of data
    DataRecord record = recordManager.getNewRecord(MAX_RECORD_SIZE);
    record.length = -1;
    try {
      dataBuffer.put(record);
      closeSync.acquire();  // block until the sending thread is committed to termination
      state = CLOSED;
    } catch (InterruptedException e) {
    }
  }

}

